@@ -1,5 +1,9 @@
 #!/usr/bin/perl
 
+BEGIN {
+	unshift @INC, 'blib/lib', 'blib/arch';
+}
+
 use threads;
 use threads::shared;
 
